#include "X2Joint.h"
// EX-106+ Manual: http://support.robotis.com/en/product/dynamixel/ex_series/ex-106.htm

X2Joint::X2Joint(int id, int PID_P, int PID_I, int PID_D) {
	kp = PID_P;
	ki = PID_I;
	kd = PID_D;

	spd_kp = 7;
	spd_ki = 10;

	servo_id = id;
	rw_pin = 2;
	
	/* Assumes the robot is in the upright vertical position */
	angle_curr = 0.0;
	angle_target = 0.0;
	//angle_maxspeed = 0.15;
	spd_angle_maxspeed = 0.075;
}


void X2Joint::beginComm() {
	//Opens the necessary communication interfaces for all joints
	pinMode(rw_pin, OUTPUT);
	Serial1.begin(1000000);
}


void X2Joint::begin() {
	/* Assumes the robot is in the upright vertical position */
	angle_curr = 0.0;
	angle_target = 0.0;
	spd_correction = 0;
	
	delay(100);
	torqueEnable = 1;
	setTorque(0,0);

	//Fetch the initial value of the servo's encoder
	while(query() == 0x3FFF) {
		//Read until we get an intial value that makes sense
		delay(100);
	}
	PID_angle_last = angle_curr;
	PID_ITerm = 0;
	spd_ITerm = 0;
}

void X2Joint::setPID(double P, double I, double D) {
	kp = P;
	ki = I;
	kd = D;
}


void X2Joint::setAngle(double angle) {
	//Sets the desired target joint angle

	//Sanity checks
	if (angle > 120) angle = 120;
	else if (angle < -120) angle = -120;

	angle_target = angle;
}


double X2Joint::getAngle() {
	return angle_curr;
}

void X2Joint::setMaxspeed(double angleper10ms) {
	spd_angle_maxspeed = angleper10ms;
}

void X2Joint::PID(int angleCompensation) {
  //Compute all the working error variables
  double error = angle_target - angle_curr;
  //PID_ITerm += (ki * error);
  PID_ITerm = (ki * error) + PID_ITerm*0.9;
  /*if(PID_ITerm > 1023) PID_ITerm = 1023;
  else if(PID_ITerm < -1023) PID_ITerm = -1023;*/
  if(PID_ITerm > 255) PID_ITerm = 255;
  else if(PID_ITerm < -255) PID_ITerm = -255;
  double dInput = (angle_curr - PID_angle_last);

  //Compute PID Output
  torque = kp * error + PID_ITerm - kd * dInput + angleCompensation;
  
  if(torque > 1023) torque = 1023;
  else if(torque < -1023) torque = -1023;

 /* double pTerm, dTerm, iTerm;
  double error = angle_target - angle_curr;

  pTerm = kp*error;
  PID_ITerm += error;
  if(PID_ITerm > 1023) PID_ITerm = 1023;
  else if(PID_ITerm < -1023) PID_ITerm = -1023;
  iTerm = ki*PID_ITerm;
  dTerm = kd*(angle_curr - PID_angle_last);

  torque = pTerm + iTerm - dTerm;
  if(torque > 1023) torque = 1023;
  else if(torque < -1023) torque = -1023;*/
  
  /*Remember some variables for next time*/
  PID_angle_last = angle_curr;
}

void X2Joint::speedLimiter(unsigned long time_curr) {
	/*if (spd_angle_sum < 0) {
		spd_angle_sum = -spd_angle_sum; //abs
	}*/

	/*if (time_curr < time_prev) {
		//timer overflow. Skip speed control this step for simplicity
		time_prev = time_curr;
		return;
	}*/

	//Time normalize the step:
	//maxspeed_angle_sum /= ((time_curr - time_prev)/10);

	//Angle_step should now contain the aproximate movement over 1 ms.
	//if (spd_angle_sum > spd_angle_maxspeed) {
		//Over max speed. Regulate with PI
		double error = spd_angle_sum - spd_angle_maxspeed; //correct direction?
		spd_ITerm = (spd_ki * error) + spd_ITerm*0.9;
		if (spd_ITerm > 1023) spd_ITerm = 1023;
		else if (spd_ITerm < -1023) spd_ITerm = -1023;

		spd_correction += spd_kp*error + spd_ITerm;

	//}/* else if (spd_divisor > 1) {
		//spd_divisor--;
	//}*/
	time_prev = time_curr;
}

void X2Joint::update(int angleCompensation) {
	double angle_step;
	//1) Asks the servo to respond with its current angle.
	//2) Fetches response
	//3) Calculates new torque value
	int last_step = query();
	if (error) {
		//Servo comm error. Skip this step's PID calculation
		//Serial.println("Skipping due to servo comm error");
		return;
	}
	angle_step = DEGREES_PER_STEP*last_step;
	angle_curr += angle_step; //Update current angle
	
	//Keep track of arm movement, but only update torque and PID values if enable is on
	if (torqueEnable) {
		PID(angleCompensation); //run one step of the PID-loop with our new value
		
		/*unsigned long time_curr = millis();
		spd_angle_sum += angle_step;
		if ((time_curr - time_prev) > 10) {
			//Calculate speed moderation factor to keep within max speed
			speedLimiter(time_curr);
			spd_angle_sum = 0;
		}
		torque += spd_correction;
		if(torque > 1023) torque = 1023;
  		else if(torque < -1023) torque = -1023;*/

		//Send newly calculated torque value to the servo.
		if (torque < 0) {
			setTorque((int)(-torque), COUNTERCLOCKWISE);
		} else {
			setTorque((int)torque, CLOCKWISE);
		}
	}
}

int X2Joint::query() {
	//Asks for and fetches this joint's servo encoder value. Returns number of steps since last query operation.
	int encoder_val = readReg2(36); //read present position registers
	if (error) {
		//Error communicating with servo. Report 0 movement.
//		Serial.println("Servo comm error");
		return 0x3FFF;
	}
	//Serial.println(encoder_val);

	//Find the step size
	int last_step = encoder_val - encoder_prev;
	if ((last_step > 30000) || (last_step < -30000)) { //Giant step detected. Probably encoder rollover
		if (encoder_prev < encoder_val) {
			//Underflow
			last_step = -encoder_prev; //Distance from 0 to encoder old
			last_step -= 65536 - encoder_val;
		} else {
			//Overflow
			last_step = encoder_val; //Distance from 0 to encoder new
			last_step += 65536 - encoder_prev; 
		}
	}
	
	encoder_prev = encoder_val;
	return last_step;
}

void X2Joint::torqueOff() {
	torqueEnable = 0; //disables PID update to prevent I-buildup until it's turned on again
	setTorque(0, 0);
	writeReg1(24, 0); //disable motor torque
}

void X2Joint::torqueOn() {
	torqueEnable = 1;
	writeReg1(24, 1); //enable motor torque
	//The actual torque is updated by the update loop next iteration
}

void X2Joint::setTorque(uint16_t value, uint8_t direction) {
	/*Serial.print("Curr: ");
	Serial.println(angle_curr);
	Serial.print("Torque: ");
	Serial.println(torque);*/

	//Sets the given torque value for this joint's servo
	value &= 0x3FF; //Clamp within 0-1023
	value |= (direction << 10); //Shift in direction bit

	//Serial.println("setTorque");

	//Write to servo
	//writeReg2(32, 0); //STOP
	writeReg2(32, value); //Write the direction bit and the 10 torque bits to the speed register
}

void X2Joint::addChecksum(uint8_t* arr, uint8_t len) {
	//Adds the checksum to the given array
	uint8_t checksum = 0;

	//Caclulate checksum
	for (uint8_t i = 2; i < len-1; i++) {
		checksum += arr[i]; //Will overflow, but that's ok
	}
	/*Serial.print("Checksum calculated: ");
	Serial.println(~checksum);*/
	arr[len-1] = ~checksum; //set checksum
}

int X2Joint::readReg2(uint8_t regLSB) { //reads two bytes starting from the specified address
	int length;
	uint8_t buf[] = {0xFF, 0xFF, servo_id, 4, 2, regLSB, 0X02, 0}; //read 2 bytes starting from regLSB

	Serial1.setTimeout(0); //0ms timeout on readBytes()

    //Throw away any pending data on the Serial1-port
    while(Serial1.available() > 0) Serial1.read();
    
	addChecksum(buf, 8);
    digitalWrite(rw_pin, HIGH);
    Serial1.write(buf, 8); //Send query
    digitalWrite(rw_pin, LOW);
    delayMicroseconds(100);

    //Reset some important fields as we don't want to react to our own query in case the read times out
    buf[2] = buf[3] = 0;
    
    //Expected packet: [0xff, 0xff, id, length, error, param1, param2, checksum]. Read the first 4:
    Serial1.readBytes(buf, 4); //may time out. CHECK RETURN VALUE TO CHECK TIMEOUT

    length = buf[3];

    error = (buf[2] != servo_id); //Check if we got a response from the right servo

    //Fetch remainder of response packet
    Serial1.readBytes(buf, length); //may time out. buf should now contain [error, param1, param2, checksum, ...]

    /*for (int i = 0; i < length-1; i++) {
    	checksum += buf[i];
    }*/

    //Discard packet if joint id didn't match packet id or if the error field is set
    error = (error || buf[0]); //error is global, check after return of function

    //buf[0-length] now contains error, parameter1, parameter2, checksum
    return ((buf[2] << 8) | buf[1]);
}

void X2Joint::writeReg2(uint8_t regLSB, uint16_t value) {
	//Write 2 bytes starting at the given servo register
	uint8_t buf[] = {0xFF, 0xFF, servo_id, 5, 3, regLSB, value & 0xff, (value >> 8), 0};
	addChecksum(buf, 9);

	digitalWrite(rw_pin, HIGH);
	Serial1.write(buf, 9);
	digitalWrite(rw_pin, LOW);
}

void X2Joint::writeReg1(uint8_t regNo, uint8_t value) {
	//Write 1 byte to the given servo register
	uint8_t buf[] = {0xFF, 0xFF, servo_id, 4, 3, regNo, value, 0};

    addChecksum(buf, 8);
    
    digitalWrite(rw_pin, HIGH);
	Serial1.write(buf, 8);
	digitalWrite(rw_pin, LOW);
}